import json
import sqlite3
from enum import Enum, unique
from typing import Any, Dict, List

import jmespath
from datayoga.blocks.jmespath_custom_functions import JmespathCustomFunctions


@unique
class Language(Enum):
    JMESPATH = "jmespath"
    SQL = "sql"


class Expression():
    def compile(self, expression: str):
        """Compiles an expression

        Args:
            expression (str): expression
        """
        pass

    def search(self, data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Executes the expression on a given data

        Args:
            data (List[Dict[str, Any]]): Data

        Returns:
            List[Dict[str, Any]]: Transformed data
        """
        pass


class SQLExpression(Expression):
    def compile(self, expression: str):
        self.conn = sqlite3.connect(":memory")
        self.expression = expression

    def filter(self, data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Tests a where clause for an SQL statement

        Args:
            data (List[Dict[str, Any]]): Data

        Returns:
            List[Dict[str, Any]]: Filtered data
        """
        # use a CTE to create the in memory data structure
        cte_clause = self._get_cte(data)

        column_names = data[0].keys()
        # fetch the CTE and bind the variables
        data_values = [row.get(colname) for row in data for colname in column_names]
        self.conn.row_factory = sqlite3.Row

        cursor = self.conn.execute(
            f"{cte_clause} select * from data where {self.expression}", data_values
        )
        return [dict(row) for row in cursor.fetchall()]

    def _get_cte(self, data: List[Any]) -> str:
        # builds a CTE expression for fetching in memory data

        column_names = data[0].keys()
        columns_clause = ','.join(f"`{col}`" for col in column_names)

        # values in the form of (?,?), (?,?)
        values_clause_row = f"({','.join('?'*len(column_names))})"
        values_clause = ','.join([values_clause_row]*len(data))

        # use a CTE to create the in memory data structure
        return f"with data({columns_clause}) as (values {values_clause})"

    def search(self, data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        try:
            fields = json.loads(self.expression)
            new_data = {}
            for field in fields:
                new_data[field] = self.exec_sql(data, fields[field])
            return new_data
        except:
            return self.exec_sql(data, self.expression)

    def exec_sql(self, data: List[Dict[str, Any]], expression: str) -> List[Dict[str, Any]]:
        """Executes an SQL statement

        Args:
            data (List[Dict[str, Any]]): Data

        Returns:
            List[Dict[str, Any]]: Query result
        """
        # use a CTE to create the in memory data structure
        data_inner = data if isinstance(data, list) else [data]
        cte_clause = self._get_cte(data_inner)

        column_names = data_inner[0].keys()

        # fetch the CTE and bind the variables
        data_values = [row.get(colname) for row in data_inner for colname in column_names]
        self.conn.row_factory = sqlite3.Row
        cursor = self.conn.execute(
            f"{cte_clause} select {expression} from data", data_values
        )

        return cursor.fetchone()[0]


class JMESPathExpression(Expression):
    # register custom functions
    options = jmespath.Options(custom_functions=JmespathCustomFunctions())

    def compile(self, expression: str):
        self.expression = jmespath.compile(expression)
        self.filter_expression = jmespath.compile(f"[?{expression}]")

    def filter(self, data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        return self.filter_expression.search(data, options=self.options)

    def search(self, data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        return self.expression.search(data, options=self.options)


def compile(language: Language, expression: str) -> Expression:
    """Gets a compiled expression class based on the language

    Args:
        language (Language): Language
        expression (str): Expression

    Returns:
        Expression: Expression class
    """
    if language == Language.JMESPATH.value:
        expression_class = JMESPathExpression()
    elif language == Language.SQL.value:
        expression_class = SQLExpression()

    expression_class.compile(expression)
    return expression_class



import copy
import importlib
import logging
import os
from os import path
from typing import Any, Dict, List

from jsonschema import validate

from datayoga import utils
from datayoga.block import Block
from datayoga.context import Context

logger = logging.getLogger(__name__)


class Job():
    """
    Job

    Attributes:
        steps List[Block]: List of steps
    """

    def __init__(self, job_settings: Dict[str, Any]):
        """
        Constructs a job and its blocks

        Args:
            job_settings (Dict[str, Any]): Job settings
        """
        validate(instance=job_settings, schema=utils.read_json(
            path.join(os.path.dirname(__file__), "schemas", "job.schema.json")))

        steps: List[Block] = []
        for step in job_settings["steps"]:
            block_name = step["uses"]
            module_name = f"datayoga.blocks.{block_name}.block"
            try:
                module = importlib.import_module(module_name)
            except ModuleNotFoundError:
                raise ValueError(f"Block {block_name} does not exist")

            block: Block = getattr(module, "Block")(step["with"])
            steps.append(block)

        self.steps = steps

    def transform(self, data: List[Dict[str, Any]], context: Context = None) -> List[Dict[str, Any]]:
        """
        Transforms data

        Args:
            data (List[Dict[str, Any]]): Data
            context (Context, optional): Context. Defaults to None.

        Returns:
            List[Dict[str, Any]]: Transformed data
        """
        transformed_data = copy.deepcopy(data)
        for step in self.steps:
            transformed_data = step.transform(transformed_data, context)
        return transformed_data



import logging
from typing import Any, Dict, List

from datayoga.block import Block as DyBlock
from datayoga.blocks import expression
from datayoga.context import Context

logger = logging.getLogger(__name__)


class Block(DyBlock):
    def init(self):
        logger.debug(f"Initializing {self.get_block_name()}")
        self.expression = expression.compile(self.properties["language"], self.properties["expression"])

    def run(self, data: List[Dict[str, Any]], context: Context = None) -> List[Dict[str, Any]]:
        logger.debug(f"Running {self.get_block_name()}")
        return self.expression.filter(data)



from datayoga.blocks.explode.block import Block


def test_explode():
    block = Block(
        {
            "field": "features",
            "target_field": "feature",
            "delimiter": ";"
        }
    )
    assert block.run([{"product": "BMW 316i", "features": "fast;black;midsize;classD"}]) == [
        {"product": "BMW 316i", "features": "fast;black;midsize;classD", "feature": "fast"},
        {"product": "BMW 316i", "features": "fast;black;midsize;classD", "feature": "black"},
        {"product": "BMW 316i", "features": "fast;black;midsize;classD", "feature": "midsize"},
        {"product": "BMW 316i", "features": "fast;black;midsize;classD", "feature": "classD"}
    ]



import json
from typing import Any, Dict

import yaml


def read_json(filename: str) -> Any:
    """
    Loads a filename as a JSON object

    Args:
        filename (str): JSON filename to load

    Returns:
        Any: JSON object
    """
    with open(filename, "r", encoding="utf8") as f:
        return json.load(f)


def read_yaml(filename: str) -> Dict[str, Any]:
    """
    Loads a filename as a YAML object

    Args:
        filename (str): YAML filename to load

    Raises:
        ValueError: In case of invalid YAML

    Returns:
         Dict[str, Any]: YAML python object
    """
    try:
        with open(filename, "r", encoding="utf8") as stream:
            return yaml.safe_load(stream)
    except Exception as e:
        raise ValueError(f"Malformed YAML: {e}")


def format_block_properties(properties: Dict[str, Any]) -> Dict[str, Any]:
    """Adds `fields` array with the passed properties in case it's missing

    Args:
        properties (Dict[str, Any]): properties

    Returns:
        Dict[str, Any]: formatted properties with `fields` array
    """
    return {"fields": [properties]} if not "fields" in properties else properties



import logging
import os
import sys
from os import path
from typing import Any, Dict, List

from jsonschema import validate

from datayoga import utils
from datayoga.context import Context

logger = logging.getLogger(__name__)


class Block():
    """
    Block

    Attributes:
        properties Dict[str, Any]: Block properties
    """

    def __init__(self, properties: Dict[str, Any]):
        """
        Constructs a block

        Args:
            properties (Dict[str, Any]): Block [properties]
        """
        self.properties = properties
        self.validate()
        self.init()

    def validate(self):
        """
        Validates block against its JSON Schema
        """
        json_schema_file = path.join(os.path.dirname(os.path.realpath(
            sys.modules[self.__module__].__file__)), "block.schema.json")
        logger.debug(f"validating {self.properties} against {json_schema_file}")
        validate(instance=self.properties, schema=utils.read_json(json_schema_file))

    def init(self):
        """
        Initializes block (abstract, should be implemented by the sub class)
        """
        pass

    def transform(self, data: List[Dict[str, Any]], context: Context = None) -> List[Dict[str, Any]]:
        """
        Transforms data

        Args:
            data (List[Dict[str, Any]]): Data
            context (Context, optional): Context. Defaults to None.

        Returns:
            List[Dict[str, Any]]: Transformed data
        """
        logger.debug(f"Transforming data, data before: {data}")

        transformed_data = self.run(data, context)
        logger.debug(f"Data after: {transformed_data}")

        return transformed_data

    def run(self, data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """ Transforms data (abstract, should be implemented by the sub class)

        Args:
            data (List[Dict[str, Any]]): Data

        Returns:
            List[Dict[str, Any]]: Transformed data
        """
        pass

    def get_block_name(self):
        return self.__class__.__name__



from datayoga.blocks.add_field.block import Block


def test_add_field():
    block = Block({"field": "full_name",
                   "language": "jmespath",
                   "expression": "[fname,lname] | join(' ', @)"})
    assert block.run([{"fname": "john", "lname": "doe"}]) == [
        {"fname": "john", "lname": "doe", "full_name": "john doe"}]



import string

from jmespath import functions


# custom functions for Jmespath
class JmespathCustomFunctions(functions.Functions):

    @functions.signature({"types": ["string"]})
    def _func_capitalize(self, arg):
        return string.capwords(str(arg))

    @functions.signature({"types": ["array"]})
    def _func_concat(self, elements):
        return ''.join([str(x) for x in elements])

    @functions.signature({"types": ["string"]})
    def _func_lower(self, element):
        return str(element).lower()

    @functions.signature({"types": ["string"]})
    def _func_upper(self, element):
        return str(element).upper()



import logging
from typing import Any, Dict, List

from datayoga.context import Context
from datayoga.job import Job

logger = logging.getLogger(__name__)


def compile(job_settings: Dict[str, Any]) -> Job:
    """
    Compiles a job in YAML 

    Args:
        job_settings (Dict[str, Any]): Job settings

    Returns:
        Job: Compiled job
    """
    logger.debug("Compiling job")
    return Job(job_settings)


def validate(job_settings: Dict[str, Any]):
    """
    Validates a job in YAML 

    Args:
        job_settings (Dict[str, Any]): Job settings

    Raises:
        ValueError: when the job is invalid
    """
    logger.debug("Validating job")
    try:
        Job(job_settings)
    except Exception as e:
        raise ValueError(e)


def transform(job_settings: Dict[str, Any],
              data: List[Dict[str, Any]],
              context: Context = None) -> List[Dict[str, Any]]:
    """
    Transforms data against a certain job

    Args:
        job_settings (Dict[str, Any]): Job settings
        data (List[Dict[str, Any]]): Data to transform
        context (Context, optional): Context. Defaults to None.

    Returns:
        List[Dict[str, Any]]: Transformed data
    """
    job = compile(job_settings)
    logger.debug("Transforming data")
    return job.transform(data, context)



import logging
from typing import Any, Dict, List

from datayoga.block import Block as DyBlock
from datayoga.context import Context

logger = logging.getLogger(__name__)


class Block(DyBlock):
    def init(self):
        logger.debug(f"Initializing {self.get_block_name()}")
        self.target_field = self.properties.get("target_field", self.properties.get("field"))
        self.delimiter = self.properties.get("delimiter", ",")
        self.field = self.properties["field"]

    def run(self, data: List[Dict[str, Any]], context: Context = None) -> List[Dict[str, Any]]:
        logger.debug(f"Running {self.get_block_name()}")
        return_data = []
        for row in data:
            # explode the field
            split_values = row[self.properties["field"]].split(self.properties["delimiter"])
            # add a list of the items with the new field
            return_data.extend([dict({self.target_field: value}, **row,) for value in split_values])

        return return_data



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['datayoga',
 'datayoga.blocks',
 'datayoga.blocks.add_field',
 'datayoga.blocks.add_field.tests',
 'datayoga.blocks.explode',
 'datayoga.blocks.explode.tests',
 'datayoga.blocks.filter',
 'datayoga.blocks.filter.tests',
 'datayoga.blocks.map',
 'datayoga.blocks.remove_field',
 'datayoga.blocks.rename_field']

package_data = \
{'': ['*'], 'datayoga': ['schemas/*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'jmespath>=1.0.0,<2.0.0', 'jsonschema>=4.4.0,<5.0.0']

extras_require = \
{'test': ['mock>=4.0.3,<5.0.0',
          'pytest>=7.1.2,<8.0.0',
          'pytest-describe>=2.0.1,<3.0.0',
          'pytest-mock>=3.7.0,<4.0.0',
          'pytest-timeout>=2.1.0,<3.0.0',
          'requests-mock>=1.9.3,<2.0.0']}

setup_kwargs = {
    'name': 'datayoga',
    'version': '0.17.0',
    'description': 'DataYoga for Python',
    'long_description': '# datayoga-py\n\n## Introduction\n\n`datayoga-py` is the transformation engine used in `DataYoga`, a framework for building and generating data pipelines.\n\n## Installation\n\n```bash\npip install datayoga\n```\n\n## Quick Start\n\nThis demonstrates how to transform data using a DataYoga job.\n\n### Create a Job\n\nUse this `example.yaml`:\n\n```yaml\nsteps:\n  - uses: add_field\n    with:\n      fields:\n        - field: full_name\n          language: jmespath\n          expression: \'{ "fname": fname, "lname": lname} | join(\'\' \'\', values(@))\'\n        - field: country\n          language: sql\n          expression: country_code || \' - \' || UPPER(country_name)\n  - uses: rename_field\n    with:\n      fields:\n        - from_field: fname\n          to_field: first_name\n        - from_field: lname\n          to_field: last_name\n  - uses: remove_field\n    with:\n      fields:\n        - field: credit_card\n        - field: country_name\n        - field: country_code\n  - uses: map\n    with:\n      expression:\n        {\n          first_name: first_name,\n          last_name: last_name,\n          greeting: "\'Hello \' || CASE WHEN gender = \'F\' THEN \'Ms.\' WHEN gender = \'M\' THEN \'Mr.\' ELSE \'N/A\' END || \' \' || full_name",\n          country: country,\n          full_name: full_name\n        }\n      language: sql\n```\n\n### Transform Data Using `datayoga-py`\n\nUse this code snippet to transform a data record using the job defined [above](#create-a-job):\n\n```python\nimport datayoga as dy\nfrom datayoga.job import Job\nfrom datayoga.utils import read_yaml\n\njob_settings = read_yaml("example.yaml")\njob = dy.compile(job_settings)\n\nassert job.transform({"fname": "jane", "lname": "smith", "country_code": 1, "country_name": "usa", "credit_card": "1234-5678-0000-9999", "gender": "F"}) == {"first_name": "jane", "last_name": "smith", "country": "1 - USA", "full_name": "jane smith", "greeting": "Hello Ms. jane smith"}\n```\n\nAs can be seen, the record has been transformed based on the job:\n\n- `fname` field renamed to `first_name`.\n- `lname` field renamed to `last_name`.\n- `country` field added based on an SQL expression.\n- `full_name` field added based on a [JMESPath](https://jmespath.org/) expression.\n- `greeting` field added based on an SQL expression.\n\n### Examples\n\n- Add a new field `country` out of an SQL expression that concatenates `country_code` and `country_name` fields after upper case the later:\n\n  ```yaml\n  uses: add_field\n  with:\n    field: country\n    language: sql\n    expression: country_code || \' - \' || UPPER(country_name)\n  ```\n\n- Rename `fname` field to `first_name` and `lname` field to `last_name`:\n\n  ```yaml\n  uses: rename_field\n  with:\n    fields:\n      - from_field: fname\n        to_field: first_name\n      - from_field: lname\n        to_field: last_name\n  ```\n\n- Remove `credit_card` field:\n\n  ```yaml\n  uses: remove_field\n  with:\n    field: credit_card\n  ```\n\nFor a full list of supported block types [see reference](https://datayoga-io.github.io/datayoga-py/).\n\n## Expression Language\n\nDataYoga supports both SQL and [JMESPath](https://jmespath.org/) expressions. JMESPath are especially useful to handle nested JSON data, while SQL is more suited to flat row-like structures.\n\n### JMESPath Custom Functions\n\nDataYoga adds the following custom functions to the standard JMESPath library:\n\n| Function     | Description                                                             | Example                                                                                                         | Comments                                                                               |\n| ------------ | ----------------------------------------------------------------------- | --------------------------------------------------------------------------------------------------------------- | -------------------------------------------------------------------------------------- |\n| `capitalize` | Capitalizes all the words in the string                                 | Input: `{name: "john doe"}` <br /> Expression: `capitalize(name)` <br /> Output: `John Doe`                     |\n| `concat`     | Concatenates an array of variables or literals                          | Input: `{fname: "john", lname: "doe"}` <br /> Expression: `concat([fname,\' \',lname])` <br /> Output: `john doe` | This is equivalent to the more verbose built-in expression: `\' \'.join([fname,lname])`. |\n| `lower`      | Converts all uppercase characters in a string into lowercase characters | Input: `{fname: "John"}` <br /> Expression: `lower(fname)` <br /> Output: `john`                                |\n| `upper`      | Converts all lowercase characters in a string into uppercase characters | Input: `{fname: "john"}` <br /> Expression: `upper(fname)` <br /> Output: `JOHN`                                |\n',
    'author': 'DataYoga',
    'author_email': 'admin@datayoga.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
from typing import Any, Dict, List

from datayoga import utils
from datayoga.block import Block as DyBlock
from datayoga.blocks import expression
from datayoga.context import Context

logger = logging.getLogger(__name__)


class Block(DyBlock):
    def init(self):
        logger.debug(f"Initializing {self.get_block_name()}")
        self.properties = utils.format_block_properties(self.properties)

        self.fields = {}
        for property in self.properties["fields"]:
            self.fields[property["field"]] = expression.compile(
                property["language"],
                property["expression"])

    def run(self, data: List[Dict[str, Any]], context: Context = None) -> List[Dict[str, Any]]:
        logger.debug(f"Running {self.get_block_name()}")

        for row in data:
            for field in self.fields:
                row[field] = self.fields[field].search(row)

        return data



from typing import Any, Dict


class Context():
    """
    Context

    Attributes:
        properties Dict[str, Any]: Context properties
        state (Dict[str, Any]): Context state
    """

    def __init__(self, properties: Dict[str, Any], state: Dict[str, Any]):
        """
        Constructs a context

        Args:
            properties (Dict[str, Any]): Context properties
            state (Dict[str, Any]): Context state
        """
        self.properties = properties
        self.state = state
